"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GolangGenerator = void 0;
var templates = __importStar(require("./templates"));
var liquidjs_1 = require("liquidjs");
var graphql_1 = require("graphql");
/**
 * GolangGenerator is a class that you instanciate with a GraphQL schema
 * and an optional configuration. You can then call its method `generate`
 * to generate Golang code.
 */
var GolangGenerator = /** @class */ (function () {
    function GolangGenerator(schema, config) {
        var _this = this;
        /**
         * Map GraphQL type to their Golang format.
         */
        this.types = {
            Int: 'Int',
            Float: 'Float',
            Boolean: 'Boolean',
            String: 'String',
            ID: 'ID',
        };
        /**
         * Map GraphQL scalar types to Golang types.
         */
        this.scalars = {
            Int: 'int32',
            Float: 'float64',
            Boolean: 'bool',
            String: 'string',
            ID: 'string',
        };
        this.enums = [];
        this.inputs = [];
        this.objects = [];
        this.config = config !== null && config !== void 0 ? config : {};
        this.schema = schema;
        this.templateEngine = new liquidjs_1.Liquid({
            strictFilters: true,
            strictVariables: true,
        });
        this.templateOperation = this.templateEngine.parse(templates.GOLANG_OPERATION);
        graphql_1.visit(graphql_1.parse(graphql_1.printSchema(this.schema)), {
            ScalarTypeDefinition: function (node) {
                _this.types[node.name.value] = _this.formatName(node.name.value);
                _this.scalars[node.name.value] = 'string';
                return false;
            },
            EnumTypeDefinition: function (node) {
                _this.types[node.name.value] = _this.formatName(node.name.value);
                _this.enums.push(node);
                return false;
            },
            InputObjectTypeDefinition: function (node) {
                _this.types[node.name.value] = _this.formatName(node.name.value);
                _this.inputs.push(node);
                return false;
            },
            ObjectTypeDefinition: function (node) {
                _this.types[node.name.value] = _this.formatName(node.name.value);
                _this.objects.push(node);
                return false;
            },
        });
    }
    GolangGenerator.prototype.generate = function (documents) {
        var _this = this;
        var l = __spreadArray(__spreadArray(__spreadArray([], this.generatePackage()), this.generateImports()), [
            templates.GOLANG_BASE,
        ]);
        documents === null || documents === void 0 ? void 0 : documents.forEach(function (document) { return l.push.apply(l, _this.generateOperations(document)); });
        l.push.apply(l, this.generateSchema());
        return l.join('\n');
    };
    /**
     * Format a name for Golang.
     * @param name Name to format.
     */
    GolangGenerator.prototype.formatName = function (name) {
        if (name.match(/^id$/i) || name.match(/^uuid$/i)) {
            return name.toUpperCase();
        }
        return name
            .replace(/(^_|_$)/, '')
            .split('_')
            .map(function (word) { return word[0].toUpperCase() + word.substr(1); })
            .join('');
    };
    /**
     * Generate Golang package header.
     */
    GolangGenerator.prototype.generatePackage = function () {
        var _a;
        return [
            "package " + ((_a = this.config.packageName) !== null && _a !== void 0 ? _a : 'graphql'),
            '',
            '// Code generated by graphql-codegen-golang ; DO NOT EDIT.',
            '',
        ];
    };
    /**
     * Generate Golang imports for operations.
     */
    GolangGenerator.prototype.generateImports = function () {
        return [
            'import (',
            '  "bytes"',
            '  "encoding/json"',
            '  "fmt"',
            '  "io/ioutil"',
            '  "net/http"',
            '  "strings"',
            ')',
            '',
        ];
    };
    /**
     * Generate a named comment section.
     * @param name Name of the section.
     */
    GolangGenerator.prototype.generateSection = function (name) {
        return ['', '//', "// " + name, '//', ''];
    };
    /**
     * Generate the Golang types from schema.
     */
    GolangGenerator.prototype.generateSchema = function () {
        return __spreadArray(__spreadArray(__spreadArray(__spreadArray([], this.generateScalars()), this.generateEnums()), this.generateInputs()), this.generateObjects());
    };
    GolangGenerator.prototype.generateScalars = function () {
        var _this = this;
        var l = __spreadArray([], this.generateSection('Scalars'));
        Object.entries(this.scalars).map(function (_a) {
            var gqlType = _a[0], goType = _a[1];
            l.push("type " + _this.types[gqlType] + " " + goType);
        });
        return l;
    };
    GolangGenerator.prototype.generateEnums = function () {
        var _this = this;
        var l = __spreadArray([], this.generateSection('Enums'));
        this.enums.forEach(function (node) {
            var _a;
            var goType = _this.types[node.name.value];
            l.push('', "type " + _this.types[node.name.value] + " string", 'const (');
            (_a = node.values) === null || _a === void 0 ? void 0 : _a.forEach(function (value) {
                var name = value.name.value;
                l.push("  " + goType + _this.formatName(name) + " " + goType + " = \"" + name + "\"");
            });
            l.push(')');
        });
        return l;
    };
    GolangGenerator.prototype.generateInputs = function () {
        var _this = this;
        var l = __spreadArray([], this.generateSection('Inputs'));
        this.inputs.forEach(function (node) {
            var _a;
            var goType = _this.types[node.name.value];
            l.push('', "type " + goType + " struct {");
            (_a = node.fields) === null || _a === void 0 ? void 0 : _a.forEach(function (field) {
                l.push(_this.generateField(field.name.value, field.type));
            });
            l.push('}');
        });
        return l;
    };
    GolangGenerator.prototype.generateObjects = function () {
        var _this = this;
        var l = __spreadArray([], this.generateSection('Objects'));
        this.objects.forEach(function (node) {
            var _a;
            var goType = _this.types[node.name.value];
            l.push('', "type " + goType + " struct {");
            (_a = node.fields) === null || _a === void 0 ? void 0 : _a.forEach(function (field) {
                l.push(_this.generateField(field.name.value, field.type));
            });
            l.push('}');
        });
        return l;
    };
    /**
     * Generate a Golang struct's field.
     * @param name Name of the GraphQL field.
     * @param type Type of the GraphQL field.
     */
    GolangGenerator.prototype.generateField = function (name, type) {
        return "  " + this.formatName(name) + " " + this.generateFieldType(type, name);
    };
    /**
     *
     * @param type TypeNode contains a GraphQL type.
     * @param fieldName GraphQL field's name for the Golang's json annotation.
     * @param prefix String accumulator to add to result when recursion is done.
     * @param nonNull Boolean accumulator to determine if type must be a pointer.
     */
    GolangGenerator.prototype.generateFieldType = function (type, fieldName, prefix, nonNull) {
        if (prefix === void 0) { prefix = ''; }
        if (nonNull === void 0) { nonNull = false; }
        if (type.kind === 'NamedType') {
            return [
                "" + (nonNull ? '' : '*') + prefix,
                this.types[type.name.value] + " ",
                "`json:\"" + fieldName + (nonNull ? '' : ',omitempty') + "\"`",
            ].join('');
        }
        if (type.kind === 'NonNullType') {
            return this.generateFieldType(type.type, fieldName, prefix, prefix !== '' ? false : true);
        }
        if (type.kind === 'ListType') {
            return this.generateFieldType(type.type, fieldName, prefix + '[]', nonNull);
        }
        throw new Error("field type \"" + type + "\" not supported!");
    };
    /**
     * Generate operations code from a GraphQL document.
     * @param document Document to get the operations from.
     */
    GolangGenerator.prototype.generateOperations = function (document) {
        var _this = this;
        var l = [];
        var typeInfo = new graphql_1.TypeInfo(this.schema);
        graphql_1.visit(document, graphql_1.visitWithTypeInfo(typeInfo, {
            FragmentSpread: {
                enter: function (field) {
                    var name = _this.formatName(field.name.value);
                    l.push(" " + name + " ");
                },
            },
            FragmentDefinition: {
                enter: function (field) {
                    var _a;
                    var name = _this.formatName(field.name.value);
                    var w = ["type  " + name + " "];
                    if (field.selectionSet) {
                        var outputType = (_a = typeInfo.getType()) === null || _a === void 0 ? void 0 : _a.toString();
                        if (outputType === null || outputType === void 0 ? void 0 : outputType.startsWith('[')) {
                            if (!outputType.endsWith('!')) {
                                w.push('*');
                            }
                            w.push('[]');
                        }
                        w.push('struct {');
                    }
                    else {
                        w.push("string `json:\"" + field.name.value + "\"`");
                    }
                    l.push(w.join(''));
                },
                leave: function (field) {
                    if (field.selectionSet) {
                        l.push("}");
                    }
                },
            },
            OperationDefinition: {
                enter: function (operation) {
                    // Anonymous operation are not supported: skip them.
                    if (!operation.name)
                        return false;
                    // Subscription operations are not supported yes.
                    if (operation.operation === 'subscription')
                        return false;
                    var name = _this.formatName(operation.name.value);
                    l.push.apply(l, _this.generateSection("" + graphql_1.print(operation).split('{', 1)[0].trim()));
                    // Generate operation variables type if any
                    l.push.apply(l, _this.generateOperationVariables(name, operation.variableDefinitions));
                    // Finally, operation's response type are generated when visiting
                    // inner AST nodes
                    l.push("type " + name + "Response struct {");
                },
                leave: function (operation) {
                    if (!operation.name)
                        return;
                    // Inner nodes have been visited. Time to close response type.
                    l.push('}', '');
                    var name = _this.formatName(operation.name.value);
                    var hasVariables = _this.generateOperationVariables(name, operation.variableDefinitions).length > 0;
                    // Generate everything except variables and response type
                    l.push.apply(l, _this.generateOperationCode(name, graphql_1.print(operation), hasVariables));
                },
            },
            Field: {
                enter: function (field) {
                    var _a;
                    var name = _this.formatName(field.name.value);
                    var w = ["  " + name + " "];
                    if (field.selectionSet) {
                        var outputType = (_a = typeInfo.getType()) === null || _a === void 0 ? void 0 : _a.toString();
                        if (outputType === null || outputType === void 0 ? void 0 : outputType.startsWith('[')) {
                            if (!outputType.endsWith('!')) {
                                w.push('*');
                            }
                            w.push('[]');
                        }
                        w.push('struct {');
                    }
                    else {
                        w.push("string `json:\"" + field.name.value + "\"`");
                    }
                    l.push(w.join(''));
                },
                leave: function (field) {
                    if (field.selectionSet) {
                        l.push("} `json:\"" + field.name.value + "\"`");
                    }
                },
            },
        }));
        return l;
    };
    /**
     * Generate Golang `type ${name}Variables struct {...}`.
     * @param name Name of the operation. Must be formatted for Golang.
     * @param variableDefinitions GraphQL variables to generate Golang fields from.
     */
    GolangGenerator.prototype.generateOperationVariables = function (name, variableDefinitions) {
        var _this = this;
        if (!variableDefinitions || variableDefinitions.length == 0) {
            return [];
        }
        var l = ["type " + name + "Variables struct {"];
        variableDefinitions.forEach(function (variable) {
            l.push(_this.generateField(variable.variable.name.value, variable.type));
        });
        l.push('}', '');
        return l;
    };
    /**
     * Generates all Golang code for an operation except `type ${name}Variables`
     * and `type ${name}Response`.
     * @param name Name of the operation. Must be formatted for Golang.
     * @param operation String version of the operation.
     * @param hasVariables Whether the operation uses a `${name}Variables` or not.
     */
    GolangGenerator.prototype.generateOperationCode = function (name, operation, hasVariables) {
        var l = [];
        l.push(this.templateEngine.renderSync(this.templateOperation, {
            name: name,
            operation: operation,
            hasVariables: hasVariables,
        }));
        return l;
    };
    return GolangGenerator;
}());
exports.GolangGenerator = GolangGenerator;
