export declare const GOLANG_BASE = "\ntype Client struct {\n\t*http.Client\n\tUrl string\n}\n\n// NewClient creates a GraphQL client ready to use.\nfunc NewClient(url string) *Client {\n\treturn &Client{\n\t\tClient: &http.Client{},\n\t\tUrl:    url,\n\t}\n}\n\ntype GraphQLOperation struct {\n\tQuery         string                 `json:\"query\"`\n\tOperationName string                 `json:\"operationName,omitempty\"`\n\tVariables     json.RawMessage        `json:\"variables,omitempty\"`\n}\n\ntype GraphQLResponse struct {\n\tData   json.RawMessage `json:\"data,omitempty\"`\n\tErrors []GraphQLError  `json:\"errors,omitempty\"`\n}\n\ntype GraphQLError map[string]interface{}\n\nfunc (err GraphQLError) Error() string {\n\treturn fmt.Sprintf(\"graphql: %v\", map[string]interface{}(err))\n}\n\nfunc (resp *GraphQLResponse) Error() string {\n\tif len(resp.Errors) == 0 {\n\t\treturn \"\"\n\t}\n\terrs := strings.Builder{}\n\tfor _, err := range resp.Errors {\n\t\terrs.WriteString(err.Error())\n\t\terrs.WriteString(\"\\n\")\n\t}\n\treturn errs.String()\n}\n\nfunc execute(client *http.Client, req *http.Request) (*GraphQLResponse, error) {\n\tif client == nil {\n\t\tclient = http.DefaultClient\n\t}\n\tresp, err := client.Do(req)\n\tif err != nil {\n\t\treturn nil, err\n\t}\n\tbody, err := ioutil.ReadAll(resp.Body)\n\tresp.Body.Close()\n\tif err != nil {\n\t\treturn nil, err\n\t}\n\treturn unmarshalGraphQLReponse(body)\n}\n\nfunc unmarshalGraphQLReponse(b []byte) (*GraphQLResponse, error) {\n\tresp := GraphQLResponse{}\n\tif err := json.Unmarshal(b, &resp); err != nil {\n\t\treturn nil, err\n\t}\n\tif len(resp.Errors) > 0 {\n\t\treturn &resp, &resp\n\t}\n\treturn &resp, nil\n}\n";
export declare const GOLANG_OPERATION = "\ntype {{name}}Request struct {\n  *http.Request\n}\n\n{% if hasVariables -%}\nfunc New{{name}}Request(url string, vars *{{name}}Variables) (*{{name}}Request, error) {\n  variables, err := json.Marshal(vars)\n  if err != nil {\n    return nil, err\n  }\n  b, err := json.Marshal(&GraphQLOperation{\n    Variables: variables,\n{%- else %}\nfunc New{{name}}Request(url string) (*{{name}}Request, error) {\n  b, err := json.Marshal(&GraphQLOperation{\n{%- endif %}\n    Query: `{{operation}}`,\n  })\n  if err != nil {\n    return nil, err\n  }\n  req, err := http.NewRequest(http.MethodPost, url, bytes.NewReader(b))\n  if err != nil {\n    return nil, err\n  }\n  req.Header.Set(\"Content-Type\", \"application/json\")\n  return &{{name}}Request{req}, nil\n}\n\nfunc (req *{{name}}Request) Execute(client *http.Client) (*{{name}}Response, error) {\n  resp, err := execute(client, req.Request)\n  if err != nil {\n    return nil, err\n  }\n  var result {{name}}Response\n  if err := json.Unmarshal(resp.Data, &result); err != nil {\n    return nil, err\n  }\nreturn &result, nil\n}\n\n{% if hasVariables -%}\nfunc {{name}}(url string, client *http.Client, vars *{{name}}Variables) (*{{name}}Response, error) {\n  req, err := New{{name}}Request(url, vars)\n{%- else %}\nfunc {{name}}(url string, client *http.Client) (*{{name}}Response, error) {\n  req, err := New{{name}}Request(url)\n{%- endif %}\n  if err != nil {\n    return nil, err\n  }\n  return req.Execute(client)\n}\n\n{% if hasVariables -%}\nfunc (client *Client) {{name}}(vars *{{name}}Variables) (*{{name}}Response, error) {\n  return {{name}}(client.Url, client.Client, vars)\n{%- else %}\nfunc (client *Client) {{name}}() (*{{name}}Response, error) {\n  return {{name}}(client.Url, client.Client)\n{%- endif %}\n}\n";
